/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */




#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "liberty.h"

#include "ttsgf.h"
#include "sgfana.h"

#define MAXTRY 400

void sgfShowConsideredMoves(void);

/* 
 * Generate computer move for COLOR.
 *
 * Return the generated move in (*i, *j).
 */
  
int
genmove(int *i, int *j, int color)
{
  float val;
  int m,n;
  double t1=0., t2=0.;

  if (showtime)
    t1 = gg_gettimeofday();

  /* Prepare our table of moves considered. */
  memset(potential_moves, 0, sizeof(potential_moves));

  /* Prepare our table of move reasons. */
  clear_move_reasons();

  /* Reset all the statistics for each move. */
  stats.nodes = 0;
  stats.position_entered = 0;
  stats.position_hits    = 0;
  stats.read_result_entered = 0;
  stats.read_result_hits    = 0;

  /* Prepare matchpat for use. */
  compile_for_match();

#if HASHING
  /* Initialize things for hashing of positions. */
  hashtable_clear(movehash);
  board_to_position(p, ko_i, ko_j, /*color,*/ &(hashdata.hashpos));
  hashdata.hashval = board_hash(p, ko_i, ko_j/*, color*/);
#endif

  /* no move is found yet. */
  *i = -1;  
  *j = -1;  
  val = -1; 

  /* Set up depth values. */
  if (hurry <= 0) {
    depth          = nominal_depth - hurry;
    ko_depth       = nominal_ko_depth - hurry;
    backfill_depth = nominal_backfill_depth - hurry;
    fourlib_depth  = nominal_fourlib_depth - hurry;
  }
  else {
    depth          = nominal_depth - hurry;
    ko_depth       = nominal_ko_depth - hurry;
    backfill_depth = nominal_backfill_depth - hurry;
    fourlib_depth  = nominal_fourlib_depth - hurry + 1;
  }

  /* Find out information about the worms and dragons. */
  {
    /* Don't print reading traces during make_worms and make_dragons unless 
       the user really wants it (verbose == 3). */

    int save_verbose=verbose;
    if ((verbose==1) || (verbose==2)) --verbose;
    if (make_worms()) {
      compute_initial_influence(color, 0);
      make_dragons();
      if (mouse)
	mouse_reasons(color);
    }
    verbose=save_verbose;
  }

  /* Map the moyos for both sides. */
  make_moyo(color);

  /* Recompute the escape potential of the dragons (escape2 field).
   *
   * FIXME: Probably both this call and the make_moyo() call should go
   * into make_dragons() at some point.
   */
   
  compute_escape_potential();
  
  if (printworms) {
    show_dragons();
  }
  else if (verbose) {
    for (m=0;m<board_size;m++)
      for (n=0;n<board_size;n++)
	if ((p[m][n])
	    && (worm[m][n].origini==m) 
	    && (worm[m][n].originj==n) 
	    && (worm[m][n].cutstone)) 
	{
	  TRACE("%scutting stones found at %m\n",
		worm[m][n].cutstone==2 ? "potential " : "", m,n); 
	}
  }	  

  compute_initial_influence(color, 1);
  
  if (printmoyo) {
/*    compute_influence();*/
    print_moyo(color);
  }
   
  if (printboard) {
    showboard(printboard); 
    if (mouse && (printboard==1))
      showboard(3);
  }

  assert(stackp == 0);

  /* Try to find empty corner moves. */
  fuseki();
  assert(stackp == 0);

  /* Try to catch your stones in semeai. */
  semeai(color);
  assert(stackp == 0);
  
  /* pattern matcher */
  shapes(color);
  assert(stackp == 0);

  /* Review the move reasons and estimate move values. */
  if (review_move_reasons(i, j, &val, color)) {
    TRACE("New move generation likes %m with value %f\n", *i, *j, val);
    if (analyzerflag & ANALYZE_RECOMMENDED)
      sgfBoardText(0, *i, *j, "P");
  }
  assert(stackp == 0);
 
  /* If the move value is 6 or lower, we look for endgame patterns too. */
  if (val <= 6.0) {
    endgame_shapes(color);
    assert(stackp == 0);
    if (review_move_reasons(i, j, &val, color)) {
      TRACE("New move generation likes %m with value %f\n", *i, *j, val);
      if (analyzerflag & ANALYZE_RECOMMENDED)
	sgfBoardText(0, *i, *j, "P");
    }
    assert(stackp == 0);
  }
 
  
  /* If no move found yet, revisit any semeai and change the
   * status of the opponent group from DEAD to UNKNOWN, then 
   * run shapes and endgame_shapesagain. This may turn up a move.
   */
  if ((val < 0.0) && revise_semeai(color)) {
    shapes(color);
    endgame_shapes(color);
    if (review_move_reasons(i, j, &val, color)) {
      TRACE("Upon reconsideration new move generation likes %m with value %f\n",
	    *i, *j, val); 
      if (analyzerflag & ANALYZE_RECOMMENDED)
	sgfBoardText(0,*i,*j,"p");
    }
  }
  
  if ((val < 0.0) && fill_liberty(i, j, color)) {
    val = 1.0;
    TRACE("Filling a liberty at %m\n", *i, *j);
    if (analyzerflag & ANALYZE_RECOMMENDED)
      sgfBoardText(0, *i, *j, "L");
  }

  /* If no move is found then pass. */
  if (val < 0) {
    TRACE("I pass.\n");
    *i = board_size;
    *j = board_size;
  } else
    TRACE("genmove() recommends %m with value %f\n", *i, *j, val);
  
  /* If statistics is turned on, this is the place to show it. */
  if (showstatistics) {
    char text[120];
    sprintf(text, "Nodes:                %d\n", stats.nodes);
    sgfAddComment(0,text);
      
    printf("Nodes:                %d\n", stats.nodes);
    printf("Positions entered:    %d\n", stats.position_entered);
    printf("Position hits:        %d\n", stats.position_hits);
    printf("Read results entered: %d\n", stats.read_result_entered);
    printf("Read result hits:     %d\n", stats.read_result_hits);
  }

  if (showtime) {
    t2 = gg_gettimeofday();
    fprintf(stderr, "TIME to generate move: %f\n", t2-t1);
  }
  
  return val;
  
}  /* end genmove */




/* This is called for each move which has been considered. For
 * debugging purposes, we keep a table of all the moves we
 * have considered.
 */

void 
move_considered(int i, int j, int val)
{
  if ((val > potential_moves[i][j])) {
    potential_moves[i][j] = val;
  }
}



/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */

